﻿local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")
local L	= LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)

SpeakinSpell:PrintLoading("substitutions.lua")

-------------------------------------------------------------------------------
-- A LIST OF ALL POSSIBLE SUBSTITUTIONS (for testing purposes)
-------------------------------------------------------------------------------

SpeakinSpell.TEST_ALL_SUBS = {
	-- basic DetectedEvent data
	"name",
	"rank",
	"type",
	"key",
	"target", -- has special handling beyond the simple UnitName("target")
	"selected", -- always UnitName("target")
	"caster",
	
	-- UnitName API tests
	"player",
	"mouseover",
	"focus",
	"pet",
	
	-- Substitution Functions from the table below
	"realm",
	"playertitle", -- " of Stormwind", "Merrymaker "
	"playerfulltitle", -- "Stonarius of Stormwind", "Merrymaker Stonarius"
	"guild",
	"home",
	"zone", -- "Stormwind"
	"subzone", -- "The Canals"
	"spelllink", -- clickable
	"targetclass",
	"targetrace",
	"scenario",
	"eventtype",
	"eventtypeprefix",
	"displayname",

	-- alias for other data
	"spellname", -- same as "name"
	"spellrank", -- same as "rank"
	
	-- symbols
	"tm",
	"r",
	"c",
	
	-- numbered party member slots for UnitName API
	"party1",
	"party2",
	"party3",
	"party4",
	"party5",
	"raid1",
	"raid5",
	"raid7",
	"raid12",
	"raid40",
	"arena1",
	"arena2",
	"arena3",
	"arena4",
	"arena5",
	
	-- number party member pets
	"partypet1",
	"partypet2",
	"partypet3",
	"partypet4",
	"partypet5",
	"raidpet2", -- not doing the other raid pets
	
	-- append "target" to UnitName API using a player's name instead of unitId (requires hyphens)
	"Stonarius-target",
	"Stonarius-target-target",
	
	-- random substitutions
	"randomtaunt",
	"randomfaction",
	"randomgirl",
	"randomboy",
	"randomsilly",
	
	-- possessive
	"player's",
	"target's",
	"caster's",
	"randomtaunt's",
	"mouseover's",
	"focus'", -- this is the correct way to specify <focus'> ... NOT <focus's>
	
	-- third person
	"player|me",
	"name|me",
	"mouseover|me",
	
	-- third person possessive
	"player's|my",
	"name's|my",
	"mouseover's|my",

	-- append "target" to UnitName API
	"playertarget",
	"playertargettarget",
	"raid1target",
	"pettarget",
	"pettargettarget",
	"partypet2target",
	
	-- append "target" ... third person possessive
	"mouseovertarget's|my",
	"playertarget's|my",
	"Stonarius-target-target's|my",

	-- gender
	"player*male*female",
	"mouseover*boy*girl",
	"Stonarius-target-target*him*her",
	-- gender... third person possessive
	"mouseover's*his*her|my", -- the possessive form is actually irrelevent because "his" and "her's" are already possessive
	"mouseover*the boy*the girl|me", -- with spaces
	"Stonarius-target-target*him*her|me", -- with target appended
	
	-- slash commands are processed by another subsystem
	--"/cast foo",
	--"/ss macro test",
	--"/ss help",
	--"/ss testallsubs", -- would cause infinite recursion!
}

-------------------------------------------------------------------------------
-- SUBSTITUTION FUNCTIONS TABLE
-------------------------------------------------------------------------------

SpeakinSpell.SubstitutionFunctions = {		
["selected"] = function(de) 
	return UnitName("target")
end,
["realm"] = function(de) 
	return GetRealmName()
end,
["playertitle"] = function(de) 
	return GetTitleName( GetCurrentTitle() )
end,
["playerfulltitle"] = function(de) 
	return SpeakinSpell:GetPlayerFullTitleName()
end,
["guild"] = function(de) 
	return SpeakinSpell:GetGuildName()
end,
["home"] = function(de) 
	return GetBindLocation()
end,
["displayname"] = function(de)
	return SpeakinSpell:FormatDisplayName( de )
end,

-- auto-substitutions like MS Word would use
["tm"] = function(de) 
	return "™"	-- Trademark symbol
end,
["r"] = function(de) 
	return "®"	-- Registered Trademark symbol
end,
["c"] = function(de) 
	return "©"	-- Copyright (C)
end,

--	GetRealZoneText is subtly different from GetZoneText re: instance names
--		see http://www.wowwiki.com/API_GetZoneText
--		we want to use GetRealZoneText to use real instance names instead of subzone names
--		<subzone> will use the minimap text for instances
--	GetSubZoneText returns an empty string if you're not in a sub zone
--		GetMinimapZoneText returns the current subzone text shown on the minimap
--		the minimap always shows something = the greater region name if not in a subzone
--		so this is guaranteed to return an appropriate value
["zone"] = function(de) 
	return GetRealZoneText()
end,
["subzone"] = function(de) 
	return GetMinimapZoneText()
end,
["spelllink"] = function(de)
	return SpeakinSpell:SafeGetSpellLink(de)
end,

-- default caster and target information
["caster"] = function(de)
	return UnitName("player")
end,
["targetclass"] = function(de)
	return UnitClass( de.target )
end,
["targetrace"] = function(de)
	return UnitRace( de.target )
end,

-- aliases for other information
["spellname"] = function(de)
	return de.name
end,
["spellrank"] = function(de)
	return de.rank
end,

-- Select channel to use while in <scenario>
-- Just a sec, I am busy with <scenario>
-- "<SpellLink> is silent while in <Scenario>"
["scenario"] = function(de)
	local Scenarios = {
		Arena		= L["The Arena"],
		BG			= L["A Battleground"],
		WG			= L["Wintergrasp"],
		RaidLeader	= L["Raid Leadership Mode"],
		RaidOfficer	= L["Raid Officer Mode"],
		Raid		= L["A Raid"],
		Party		= L["A Party"],
		PartyLeader	= L["Party Leadership Mode"],
		Solo		= L["My Alone Time"],
	}
	return Scenarios[ SpeakinSpell:GetScenarioKey() ]
end,

["eventtypeprefix"] = function(de)
	--i.e. "When I Start Casting: "
	return L.EVENTTYPES.IN_SPELL_LIST[de.type]
end,

["eventtype"] = function(de)
	--i.e. "Spells, Abilities, and Items (Start Casting)"
	return L.EVENTTYPES.AS_FILTERS[de.type]
end,

["newline"] = function(de)
	return "\n"
end,

}


-- see substititions.lua for substitution lookup/generation functions
-- random substititions are handled through another system
function SpeakinSpell:LookupSubstitutionFunctionTable( sk, de )
	local Func = SpeakinSpell.SubstitutionFunctions[ sk.key ]
	if Func then
		--self:DebugMsg(nil,"found function for key:"..tostring(sk.key))
		return Func(de)
	else
		--self:DebugMsg(nil,"no function for key:"..tostring(sk.key))
		return nil
	end
end


-------------------------------------------------------------------------------
-- UNIT CHANGES FOR GRAMMAR
-------------------------------------------------------------------------------


-- Applies possessive logic to a substitution
-- if the substitution key was possessive, make the unit possessive as well
function SpeakinSpell:MakeUnitPossessive(sk)
	if sk.nonpossessive then
		sk.unitnonpossessive = sk.unit
		sk.unit = self:MakePossessive( sk.unit )
	end
end


-- Applies Third Person, Gender, or Me logic to the SubstitutionKey (sk)
-- if the unit is me, use "me" text instead of speaking in the third person
-- if gender text was specified for male and female, and the unit gender is known, use the gender text
function SpeakinSpell:PickUnitGenderOrMe(sk)
	if sk.me then
		sk.unit = sk.me
	elseif sk.unitgender and sk.male then
		if sk.unitgender == 2 and sk.male then
			sk.unit = sk.male
		elseif sk.unitgender == 3 and sk.female then
			sk.unit = sk.female
		end
	end
	-- else, sk.unit remains unchanged
end


-------------------------------------------------------------------------------
-- MAIN SEARCH FUNCTION
-------------------------------------------------------------------------------

-- takes a SubstitutionKey (sk) and a CurrentEvent and searches for the sk.key information
-- sets sk.unit to the matching value
-- grammar features are not applied
function SpeakinSpell:FindUnitForSub( sk, CurrentEvent )
	local funcname = "FindUnitForSub"
	--self:DebugMsg(funcname, "sk.key:"..sk.key)

	-- Try parent event information first
	local de = CurrentEvent
	if self.RuntimeData.ParentDetectedEvent then
		--self:DebugMsg(funcname, "Using Parent Event Info")
		de = self.RuntimeData.ParentDetectedEvent
	end
	
	-- define a search order
	local SearchOrder = {
		-- Basic Substitutions from the event information
		[1] = function( sk, de )
			return de[ sk.key ]
		end,
		-- In case the Current Event defines something the Parent Event doesn't
		[2] = function( sk, de )
			return CurrentEvent[ sk.key ]
		end,
		-- UnitName API Substitutions
		[3] = function( sk, de )
			return UnitName( sk.key )
		end,
		-- Function Table
		[4] = function( sk, de )
			return self:LookupSubstitutionFunctionTable( sk, de )
		end,
		-- Random Subtititutions
		[5] = function( sk, de )
			local last = SpeakinSpell.RuntimeData.LastRandomSub[sk.key]
			local unit = self:GetRandomTableEntry( SpeakinSpellSavedData.RandomSubs[sk.key], last )
			SpeakinSpell.RuntimeData.LastRandomSub[sk.key] = unit
			return unit
		end,
	}
	
	-- search through all the tables or functions for a value
	for index,Search in ipairs( SearchOrder ) do
		sk.unit = Search( sk, de )
		if sk.unit then
			--self:DebugMsg(funcname, "sk.key:"..sk.key.." index:"..index.." sk.unit:"..tostring(sk.unit))
			return sk.unit
		end
	end

	-- unknown -- force it to have a value
	sk.unit = "<"..sk.key..">"
	--self:DebugMsg(funcname, "unknown sk.unit:"..sk.unit)
end


-------------------------------------------------------------------------------
-- SUBSTITUTE "ME" OR "GENDER" VALUES
-------------------------------------------------------------------------------


function SpeakinSpell:SubstituteGender( sk )
	-- if gender substitutions weren't specified, then forget it
	if not (sk.male or sk.female) then
		return false
	end
	
	-- figure out the gender
	if (not sk.unitgender) or (sk.unitgender == 1) then
		-- try the actual unit name first in case it's a player
		sk.unitgender = UnitSex( sk.unit )
		-- for NPCs, we can't get the gender by name,
		-- but if it's a unitId key like "mouseover" we can get it from the substitution key
		if (not sk.unitgender) or (sk.unitgender == 1) then
			sk.unitgender = UnitSex( sk.key )
			-- if we still don't know the gender, stick with the name
			if (not sk.unitgender) or (sk.unitgender == 1) then
				return false
			end
		end
	end
	
	-- if we known the gender, and we have text for that gender, use it to override the name
	if sk.unitgender == 2 and sk.male then
		sk.unit = sk.male
		return true
	elseif sk.unitgender == 3 and sk.female then
		sk.unit = sk.female
		return true
	end
	
	-- we don't have text for this gender
	return false
end


function SpeakinSpell:SubstituteMe( sk )
	if sk.me and (sk.unit == UnitName("player")) then
		sk.unit = sk.me
		return true
	else
		return false
	end
end


-------------------------------------------------------------------------------
-- CONVERT A SINGLE SUBSTITUTION VALUE - MAIN
-------------------------------------------------------------------------------

-- 'text' is what comes between the <brackets> not including brackets
-- returns the complete substitution value, with all grammar rules applied, etc
function SpeakinSpell:DoSubstitution( text, DetectedEvent )
	local funcname = "DoSubstitution"

	-- create a SubstitutionKey object to track the parsing and substitution process
	local sk = self:ParseSubstitutionKey( text )
	
	-- Find the 'unit' to which this substitution refers
	-- this guarantee's we have a valid sk.unit
	self:FindUnitForSub( sk, DetectedEvent )
	--self:DebugMsg(funcname, "sk.unit:"..tostring(sk.unit))
	
	-- make sure we save/validate some data after that search
	sk.unit = tostring( sk.unit ) -- in case something funky happened with sub-tables
	
	-- if |me was specified, and it's me, this overrides other grammar rules below
	if self:SubstituteMe( sk ) then
		return sk.unit
	end
	
	-- if gender was specified, don't make it overly possessive like <player's*his*her's> -> his's
	if self:SubstituteGender( sk ) then
		return sk.unit
	end
	
	-- make the unit possessive if the text was possessive
	self:MakeUnitPossessive( sk )
	return sk.unit
end


-------------------------------------------------------------------------------
-- REPLACE ALL SUBSTITUTIONS IN A STRING - MAIN
-------------------------------------------------------------------------------

-- search for substitutions and replace them with values
function SpeakinSpell:FormatSubs(msg, DetectedEvent)
	--self:DebugMsg("FormatSubs",msg)
	return string.gsub(msg, "<(.-)>", 
		function(var)  --<var>
			return self:DoSubstitution( var, DetectedEvent ) 
		end
	)
end


function SpeakinSpell:TestAllSubs()
	-- set this up as a detected event
	local DetectedEventStub = {
		-- event descriptors
		name = "SpeakinSpell:TestAllSubs()",
		type = "TEST",
		-- event-specific data for substitutions
		-- None
	}
	local de = self:CreateDetectedEvent( DetectedEventStub )
	
	-- for each substitution in the test list, print...
	-- player="<player>" -- for unknowns
	-- player="Stonius" -- for knowns
	for _,sub in ipairs(SpeakinSpell.TEST_ALL_SUBS) do
		local msg = string.format([[%s="<%s>"]],sub,sub)
		msg = SpeakinSpell:FormatSubs( msg, de )
		self:Print(msg)
	end
end
